"""
Train and eval functions used in main.py
"""
import math
import sys
from typing import Iterable, Optional

import torch

from timm.data import Mixup
from timm.utils import accuracy, ModelEma

from losses import DistillationLoss
import utils
import wandb

class NativeScalerGA:
    state_dict_key = "amp_scaler"

    def __init__(self):
        self._scaler = torch.cuda.amp.GradScaler()

    def __call__(self, loss, optimizer, clip_grad=None, clip_mode='norm', parameters=None, create_graph=False, do_step=True):
        self._scaler.scale(loss).backward(create_graph=create_graph)
        if clip_grad is not None:
            assert parameters is not None
            self._scaler.unscale_(optimizer)  # unscale the gradients of optimizer's assigned params in-place
            dispatch_clip_grad(parameters, clip_grad, mode=clip_mode)
        if do_step:
            self._scaler.step(optimizer)
            self._scaler.update()

    def state_dict(self):
        return self._scaler.state_dict()

    def load_state_dict(self, state_dict):
        self._scaler.load_state_dict(state_dict)


def train_one_epoch(model: torch.nn.Module, criterion: DistillationLoss,
                    data_loader: Iterable, optimizer: torch.optim.Optimizer,
                    device: torch.device, epoch: int, loss_scaler, max_norm: float = 0,
                    model_ema: Optional[ModelEma] = None, mixup_fn: Optional[Mixup] = None,
                    set_training_mode=True, accumulation_step=1, current_epoch=0, use_wandb=True):
    model.train(set_training_mode)
    metric_logger = utils.MetricLogger(delimiter="  ")
    metric_logger.add_meter('lr', utils.SmoothedValue(window_size=1, fmt='{value:.6f}'))
    header = 'Epoch: [{}]'.format(epoch)
    print_freq = 10
    count = 0
    norm_dict = {'param_norm':{}, 'grad_norm':{}}
    for name, p in model.named_parameters():
        if p.requires_grad:
            norm_dict['param_norm']['param_norm/'+name] = norm_dict['grad_norm']['grad_norm/'+name] = None

    for samples, targets in metric_logger.log_every(data_loader, print_freq, header):
        count+=1
        do_step = (count%accumulation_step==0)
        samples = samples.to(device, non_blocking=True)
        targets = targets.to(device, non_blocking=True)

        if mixup_fn is not None:
            samples, targets = mixup_fn(samples, targets)

        with torch.cuda.amp.autocast():
            outputs = model(samples)
            loss = criterion(samples, outputs, targets)

        loss_value = loss.item()

        if not math.isfinite(loss_value):
            print("Loss is {}, stopping training".format(loss_value))
            sys.exit(1)
        
        if count != 1:
            for name, p in model.named_parameters():
                if p.requires_grad:
                    norm_dict['param_norm']['param_norm/'+name] = p.norm()
                    norm_dict['grad_norm']['grad_norm/'+name] = p.grad.norm()

        if utils.is_main_process() and use_wandb: # not more ouput
            if (count-1) % print_freq == 0 and count != 1:
                for value in norm_dict.values():
                    wandb.log(value) 

        if do_step:
            optimizer.zero_grad()

        # this attribute is added by timm on one optimizer (adahessian)
        is_second_order = hasattr(optimizer, 'is_second_order') and optimizer.is_second_order
        loss_scaler(loss, optimizer, clip_grad=max_norm,
                    parameters=model.parameters(), create_graph=is_second_order, do_step=do_step)

        torch.cuda.synchronize()
        if model_ema is not None and do_step:
            model_ema.update(model)

        metric_logger.update(loss=loss_value)
        metric_logger.update(lr=optimizer.param_groups[0]["lr"])
        # add for wandb
        metrics = { "train/step": count + 1, 
                    "train/epoch": current_epoch,
                    "train/learning_rate": optimizer.param_groups[0]["lr"],
                    "train/train_loss": loss_value,
        }

        if utils.is_main_process() and use_wandb: # not more ouput
            if (count+1) % print_freq == 0:
                wandb.log(metrics)
    # gather the stats from all processes
    metric_logger.synchronize_between_processes()
    print("Averaged stats:", metric_logger)
    return {k: meter.global_avg for k, meter in metric_logger.meters.items()}


@torch.no_grad()
def evaluate(data_loader, model, device, use_wandb=True):
    criterion = torch.nn.CrossEntropyLoss()

    metric_logger = utils.MetricLogger(delimiter="  ")
    header = 'Test:'

    # switch to evaluation mode
    model.eval()

    for images, target in metric_logger.log_every(data_loader, 10, header):
        images = images.to(device, non_blocking=True)
        target = target.to(device, non_blocking=True)

        # compute output
        with torch.cuda.amp.autocast():
            output = model(images)
            loss = criterion(output, target)

        acc1, acc5 = accuracy(output, target, topk=(1, 5))

        batch_size = images.shape[0]
        metric_logger.update(loss=loss.item())
        metric_logger.meters['acc1'].update(acc1.item(), n=batch_size)
        metric_logger.meters['acc5'].update(acc5.item(), n=batch_size)
    # gather the stats from all processes
    metric_logger.synchronize_between_processes()
    print('* Acc@1 {top1.global_avg:.3f} Acc@5 {top5.global_avg:.3f} loss {losses.global_avg:.3f}'
          .format(top1=metric_logger.acc1, top5=metric_logger.acc5, losses=metric_logger.loss))
    # add for wandb
    val_metrics = {"val/val_loss": metric_logger.loss.global_avg, 
                    "val/acc@1": metric_logger.acc1.global_avg,
                    "val/acc@5": metric_logger.acc5.global_avg,
    }
    if utils.is_main_process() and use_wandb: # not more ouput
        wandb.log(val_metrics)

    return {k: meter.global_avg for k, meter in metric_logger.meters.items()}
